@@ -46,6 +46,7 @@
 #include <stdio.h>
 #include <errno.h>
 #include <string.h>
+#include <strings.h>
 #include <unistd.h>
 #include <pthread.h>
 #include <sys/types.h>
@@ -89,6 +90,16 @@ static void syserr_for_ev(const char* msg)
     log_fatal("%s: %s", msg, logf_errno());
 }
 
+static void* alloc_for_ev(void* ptr, long size)
+{
+    if (unlikely(size < 0))
+        log_fatal("Invalid alloc_for_ev() size %li\n", size);
+    if (size)
+        return xrealloc(ptr, (size_t)size);
+    free(ptr);
+    return 0;
+}
+
 static pthread_t zones_reloader_threadid;
 
 static bool join_zones_reloader_thread(void)
@@ -328,7 +339,7 @@ static void do_tak2(struct ev_loop* loop, const csc_t* csc)
             log_fatal("REPLACE[new daemon]: takeover phase 2 notification attempt failed");
         const size_t chal_count = csbuf_get_v(&resp);
         const size_t chal_dlen = resp.d;
-        log_devdebug("TAK2 challenge handoff got count %zu dlen %zu", chal_count, chal_dlen);
+        log_debug("TAK2 challenge handoff got count %zu dlen %zu", chal_count, chal_dlen);
         size_t offset = 0;
         for (size_t i = 0; i < chal_count; i++) {
             if (offset + 5U > chal_dlen)
@@ -479,12 +490,14 @@ static css_t* runtime_execute(const char* argv0, socks_cfg_t* socks_cfg, css_t*
     // Initialize dnspacket stuff
     dnspacket_global_setup(socks_cfg);
 
-    // Set up libev error callback
+    // Set up libev error+allocator callbacks
     ev_set_syserr_cb(&syserr_for_ev);
+    ev_set_allocator(&alloc_for_ev);
 
-    // default ev loop in main process to handle statio, monitors, control
-    // socket, signals, etc.
-    struct ev_loop* loop = ev_default_loop(EVFLAG_AUTO);
+    // primary ev loop in main process to handle statio, monitors, control
+    // socket, signals, etc.  Note non-default loop, meaning no implicit
+    // SIGCHLD handler from libev.
+    struct ev_loop* loop = ev_loop_new(EVFLAG_AUTO);
     if (!loop)
         log_fatal("Could not initialize the default libev loop");
 
@@ -599,6 +612,7 @@ static css_t* runtime_execute(const char* argv0, socks_cfg_t* socks_cfg, css_t*
     // as it could abort our clean shutdown sequence.
     ev_signal_stop(loop, p_sig_term);
     ev_signal_stop(loop, p_sig_int);
+    ev_loop_destroy(loop);
 
 #ifdef GDNSD_COVERTEST_EXIT
     // We have to use exit() when testing coverage, as raise()
@@ -681,7 +695,7 @@ int main(int argc, char** argv)
     }
 
     // Load full configuration and expose through the global "gcfg"
-    gcfg = conf_load(cfg_root, socks_cfg, copts.force_zsd);
+    gcfg = conf_load(cfg_root, copts.force_zsd);
     vscf_destroy(cfg_root);
 
     // Basic init for the acme challenge code
